//File: CommonBlockClasses.h
//Purpose: Define common classes to be used in a block
//world game.  These classes should not be dependent on
//any Infinifrag specific classes.  The only file
//that this should ever be dependent on is
//"Vector.h" that declares the 3D Vector class.

#ifndef CommonBlockClasses_h_
#define CommonBlockClasses_h_

//May want to switch these to be
//for a right handed coordinate system.
#define DIR_UP		1
#define DIR_DOWN	2
#define DIR_FRONT	3 //Same as south, positive y
#define DIR_BACK	4 //Same as north, negative y
#define DIR_LEFT	5 //Same as west, negative x
#define DIR_RIGHT	6 //Same as east, positive x

#include "Vector.h"

#include <cstdlib>
#include <iostream>
using std::cerr;
using std::endl;

//IMPORTANT INFORMATION:
//Here is how the world is orriented:
//The orrigon is at Vector(0, 0, 0).
//When standing at the orrigon looking north,
//the positive x-axis is EAST, the positive y-asis
//is SOUTH (because DirectX is left handed), and
//the positive z-axis is UP.
//Thus, where x, y, and z are longs, the lower
//north west corner of a cube at CubePos(x,y,z)
//is Vector(x,y,z).

//Represents the position of a block.
//All parameters should be positive
//for a block in the world.
//The types should NOT be changed to unsigned
//because otherwise loss of precision warnings
//will be generated by the compiler in routines
//that convert the unsigned to singed to see if a
//block is outside of the world.
class BlockPos {
public:
	long x, y, z;

	BlockPos() {}

	//Vanilla constructor.
	BlockPos(long x, long y, long z) {
		this->x = x;
		this->y = y;
		this->z = z;
	}

	//Arbitrary
	bool operator<(const BlockPos & b) const {
		if( x < b.x ) return true;
		if( x > b.x ) return false;
		if( y < b.y ) return true;
		if( y > b.y ) return false;
		if( z < b.z ) return true;
		return false;
	}

	BlockPos operator +(const BlockPos & b) const {
		BlockPos new_pos;
		new_pos.x = x + b.x;
		new_pos.y = y + b.y;
		new_pos.z = z + b.z;
		return new_pos;
	}
	
	BlockPos operator =(const BlockPos & b) {
		x = b.x;
		y = b.y;
		z = b.z;
		return *this;
	}

	bool operator ==(const BlockPos & b) const {
		return x == b.x && y == b.y && z == b.z;
	}
};

//Represents a 3D rectangular group of blocks in
//the world.
class BlockRect {
public:
	BlockPos min, max;

	BlockRect() {}

	//Vanilla constructor.
	BlockRect(BlockPos & min, BlockPos & max) {
		this->min = min;
		this->max = max;
	}
	
	bool operator <(const BlockRect & b) const {
		if( min < b.min ) return true;
		if( b.min < min ) return false;
		if( max < b.max ) return true;
	return false;
}
};

//Represents a 3D rectangular region in space.
class VectorRect {
public:
	Vector min, max;

	VectorRect() {}

	//Vanilla constructor.
	VectorRect(Vector & min, Vector & max) {
		this->min = min;
		this->max = max;
	}
};

//Should probably have a better name.
VectorRect inline SumRects(VectorRect & a, VectorRect & b) {
	VectorRect r;
	r.min.x = ( a.min.x < b.min.x ? a.min.x : b.min.x );
	r.min.y = ( a.min.y < b.min.y ? a.min.y : b.min.y );
	r.min.z = ( a.min.z < b.min.z ? a.min.z : b.min.z );
	r.max.x = ( a.max.x > b.max.x ? a.max.x : b.max.x );
	r.max.y = ( a.max.y > b.max.y ? a.max.y : b.max.y );
	r.max.z = ( a.max.z > b.max.z ? a.max.z : b.max.z );
	return r;
}

BlockRect inline SumBlockRects(const BlockRect & a, const BlockRect & b) {
	BlockRect r;
	r.min.x = ( a.min.x < b.min.x ? a.min.x : b.min.x );
	r.min.y = ( a.min.y < b.min.y ? a.min.y : b.min.y );
	r.min.z = ( a.min.z < b.min.z ? a.min.z : b.min.z );
	r.max.x = ( a.max.x > b.max.x ? a.max.x : b.max.x );
	r.max.y = ( a.max.y > b.max.y ? a.max.y : b.max.y );
	r.max.z = ( a.max.z > b.max.z ? a.max.z : b.max.z );
	return r;
}

inline BlockPos AdjacentBlock(const BlockPos & pos, int direction) {
	switch(direction) {
		case DIR_UP:	return BlockPos(pos.x, pos.y, pos.z+1);
		case DIR_DOWN:	return BlockPos(pos.x, pos.y, pos.z-1);
		case DIR_FRONT:	return BlockPos(pos.x, pos.y+1, pos.z);
		case DIR_BACK:	return BlockPos(pos.x, pos.y-1, pos.z);
		case DIR_LEFT:	return BlockPos(pos.x-1, pos.y, pos.z);
		case DIR_RIGHT:	return BlockPos(pos.x+1, pos.y, pos.z);
	}
	cerr << "Error: Bad direction passed to AdjacentBlock" << endl;
	exit(0);
}

inline Vector DirectionToVector(int direction) {
	//more!!!
	switch(direction) {
		case DIR_UP:	return Vector( 0.0f, 0.0f, 1.0f);
		case DIR_DOWN:	return Vector( 0.0f, 0.0f,-1.0f);
		case DIR_FRONT:	return Vector( 0.0f, 1.0f, 0.0f);
		case DIR_BACK:	return Vector( 0.0f,-1.0f, 0.0f);
		case DIR_LEFT:	return Vector(-1.0f, 0.0f, 0.0f);
		case DIR_RIGHT:	return Vector( 1.0f, 0.0f, 0.0f);
	}
	cerr << "Error: Bad direction passed to DirectionToVector" << endl;
	exit(0);
}

//Returns which block position a vector is inside of.
//Let's end the function name with a preposition.
inline BlockPos BlockInsideOf(const Vector pos) {
	long x = (long)floor(pos.x);
	long y = (long)floor(pos.y);
	long z = (long)floor(pos.z);
	return BlockPos(x,y,z);
}

//Returns true iff the block is inside of the block rect.
inline bool BlockInBlockRect(const BlockPos & block, const BlockRect & block_rect) {
	return ( block.x >= block_rect.min.x &&
			 block.y >= block_rect.min.y &&
			 block.z >= block_rect.min.z &&
			 block.x <= block_rect.max.x &&
			 block.y <= block_rect.max.y &&
			 block.z <= block_rect.max.z );
}

//Returns a BlockRect that entirely covers a VectorRect.
//If the sides of the VectorRect are integral, they will still
//be contained within the interior of the BlockRect (the
//BlockRect will just be large).
inline BlockRect CoveringBlockRect(const VectorRect & vector_rect) {
	BlockRect block_rect;
	block_rect.min.x = (long)ceil(vector_rect.min.x - 1.0f);
	block_rect.min.y = (long)ceil(vector_rect.min.y - 1.0f);
	block_rect.min.z = (long)ceil(vector_rect.min.z - 1.0f);
	block_rect.max.x = (long)floor(vector_rect.max.x + 1.0f);
	block_rect.max.y = (long)floor(vector_rect.max.y + 1.0f);
	block_rect.max.z = (long)floor(vector_rect.max.z + 1.0f);
	return block_rect;
}

#endif
